"""Update data structures code"""

from soil.data_structures.es_data_structure import ESDataStructure
from soil.data_structures.predictions import Predictions

print("updating code - !")
print(Predictions, ESDataStructure)



"""
    DataStructure to serialize a Keras Model
    It puts the network architecture as metadata and pickles the weights.
"""
import pickle

# import logging
from soil.storage.object_storage import ObjectStorage
from soil.data_structures.data_structure import DataStructure


class RandomForestModel(DataStructure):
    """Data Structure for a Sklearn Model"""

    @classmethod
    def deserialize(cls, serialized, metadata):
        """Function to deserialize"""
        return cls(pickle.loads(serialized.get_object()), metadata)

    def serialize(self):
        """Function to serialize"""
        obj_storage = ObjectStorage()
        obj_storage.put_object(pickle.dumps(self.data))
        return obj_storage

    def get_data(self, **_args):
        # pylint: disable=no-self-use
        """Placeholder function for the API call"""
        return {}



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring
# pylint: disable=no-self-use, protected-access
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch, call
# from types import GeneratorType
import pandas
from soil.storage.elasticsearch import Elasticsearch
from iris_dataset.data_structures import es_data_structure


class TestESDataStructure(unittest.TestCase):
    @patch.object(es_data_structure.ESDataStructure, "__init__", return_value=None)
    def test_deserialize_expected_inputs(self, mock_init):
        es_mock = MagicMock(spec=Elasticsearch)
        instance = es_data_structure.ESDataStructure.deserialize(es_mock, "ok2")
        assert isinstance(instance, es_data_structure.ESDataStructure)
        assert mock_init.mock_calls == [call(None, "ok2", storage=es_mock)]

    # Tests deserialize when object storage is none or not an object storage
    # DataStructure and object storage classes are required to do these tests)

    def test_deserialize_no_object_storage(self):
        self.assertRaises(
            NotImplementedError,
            es_data_structure.ESDataStructure.deserialize,
            None,
            MagicMock(spec=Elasticsearch),
        )

    def test_deserialize_no_es_storage(self):
        self.assertRaises(
            NotImplementedError,
            es_data_structure.ESDataStructure.deserialize,
            "something",
            MagicMock(spec="Invalid db"),
        )

    # def test_get_data(self):
    #     obj = es_data_structure.ESDataStructure()
    #     assert obj.get_data() == {}

    def test_serialize_no_index(self):
        obj = es_data_structure.ESDataStructure()
        self.assertRaises(AttributeError, obj.serialize)

    def test_serialize_no_data(self):
        obj = es_data_structure.ESDataStructure()
        obj.metadata = {"index": "test_index"}
        self.assertRaises(AttributeError, obj.serialize)

    @patch.object(Elasticsearch, "__init__", return_value=None)
    def test_serialize_expected(self, es_mock):
        obj = es_data_structure.ESDataStructure()
        obj.metadata = {"index": "test_index"}

        # This is to avoid an attributeError (see previous test)
        df = pandas.DataFrame({"test": [1, 2, 3, 4]}).to_dict("records")
        obj.data = df

        res = obj.serialize()
        assert es_mock.mock_calls[0] == call(index="test_index")
        assert isinstance(res, Elasticsearch)



""" A generic datastructure to index stuff to Elastic Search """
from soil.data_structures.data_structure import DataStructure
from soil.storage.elasticsearch import Elasticsearch
import soil

#########
# UTILS #
#########

MATCH_ALL = {"match_all": {}}
QUERY_ALL = {"query": MATCH_ALL}


def iris_to_es_query(filters):
    """
    Build elastic search query
    """
    must = []
    if filters is None:
        soil.logger.warning("NONE FILTERS")
    for f in filters:
        if f is None:
            return MATCH_ALL
        if "fn" not in f:
            soil.logger.warning("Query function not found!")
            return MATCH_ALL
        if f["fn"] == "list":
            must.append({"terms": {f["variable"]: f["value"]}})
        elif f["fn"] == "list-deactivated-default":
            if len(f["value"]) > 0:
                must.append({"terms": {f["variable"]: f["value"]}})
        elif f["fn"] == "equal":
            must.append({"term": {f["variable"]: f["value"]}})
        elif f["fn"] == "range":
            must.append({"range": {f["variable"]: f["value"]}})
        else:
            soil.logger.warning("Query function not found!")

    if not must:
        return {"bool": {"must": must}}
    return MATCH_ALL


def get_agg(variable):
    """
    Build an aggregation
    """
    if variable["type"] in {"terms", "avg"}:
        return {variable["name"]: {variable["type"]: {"field": variable["variable"]}}}
    return None


def iris_to_es_aggs(aggregation):
    """
    Build aggregations
    """
    if len(aggregation) == 1:
        variable = aggregation[0]
        return get_agg(variable)
    aggs = {}
    for variable in aggregation:
        aggs[variable["name"]] = get_agg(variable)[variable["name"]]

    return aggs


def get_barchart_count(hits, label="label"):
    """
    Get the barchart for counting elements in buckets
    """
    serie = []
    for d in hits["group_by_type"]["buckets"]:
        serie.append({"x": d["key"], "y": d["doc_count"]})

    if len(serie) == 0:
        return "empty"

    data = [{"serie": label, "data": serie}]
    return data


def get_barchart_sizes(hits):
    """
    Get the barchart with the measurements of the plants
    """
    data = [
        {
            "serie": "Petal",
            "data": [
                {"x": "Width", "y": hits["petal_width"]["value"]},
                {"x": "Length", "y": hits["petal_length"]["value"]},
            ],
        },
        {
            "serie": "Sepal",
            "data": [
                {"x": "Width", "y": hits["sepal_width"]["value"]},
                {"x": "Length", "y": hits["sepal_width"]["value"]},
            ],
        },
    ]
    return data


def get_hits_count(hits):
    """
    Count how many hits we have
    """
    # THIS CAN BE DONE WAY MORE EFFICIENTLY!
    return len(hits)


fns = {
    "get_barchart_sizes": get_barchart_sizes,
    "get_barchart_count": get_barchart_count,
    "get_hits_count": get_hits_count
}


##################
# DATA STRUCTURE #
##################


class ESDataStructure(DataStructure):
    # pylint:disable=attribute-defined-outside-init, access-member-before-definition
    """A generic datastructure to index stuff to Elastic Search"""

    def serialize(self):
        """
        Serializes the data and stores them using Elasticsearch.
        The data must be json serializable.
        """
        elasticsearch = Elasticsearch(index=self.metadata["index"])
        if self.metadata.get("rewrite", False):
            elasticsearch.delete_index()
        elasticsearch.create_index(schema=self.metadata.get("schema", None))
        elasticsearch.bulk(self.data)
        return elasticsearch

    @classmethod
    # The argument order is given by DataStructure!
    def deserialize(cls, es_storage, metadata):
        """
        Deserializes the data obtaining them from ObjectStorage.
        """
        if es_storage is not None and isinstance(es_storage, Elasticsearch):
            # If a datastructure comes from a database, it is prefered not to return data
            # because we do not know its size and we might want to make queries.
            # In modules, data might be accesed using get_data() or the iterator (list(ESDataStructure)).
            return cls(None, metadata, storage=es_storage)
        raise NotImplementedError("ESDataStructure without ES not implemented")

    # If we want to acces data through an interator
    def __iter__(self):
        if self.data is None:
            result = self.storage.search(
                {"query": {"match_all": {}}}, auto_scroll=True
            )  # autoscroll True the data is returned in blocks of 10000 entries
            self.data = result[0]
            self.iter = iter(self.data)
        return self

    def __next__(self):
        return next(self.iter)

    def get_data(
        self,
        query=None,
        source=None,
        aggregation=None,
        fn=None,
        fn_args=None,
        lang="ca",
    ):
        """Serialize the data for JSON the data"""

        # Return empty if there is no query
        if query is None:
            return []

        # Extract the typof query
        query_type = query["type"]

        # If the query type is iris
        if query_type == "iris":
            filters = query["filters"]
            return self.get_data_iris(
                query_type,
                filters,
                aggregation=aggregation,
                fn=fn,
                fn_args=fn_args,
                lang=lang,
                source=source,
            )

        # else
        soil.logger.warning("Query type {query_type} not implemented.")
        return []

    def get_data_iris(
        self,
        qtype,
        filters,
        aggregation=None,
        source=None,
        fn=None,
        fn_args=None,
        lang="ca",
    ):
        """
        Build and return query to IRIS front end
        """

        # Build ES query
        question = {}

        if qtype == "ES":
            question["query"] = filters

        elif qtype == "iris":
            question["query"] = iris_to_es_query(filters)

        if aggregation is not None:
            question["aggs"] = iris_to_es_aggs(aggregation)

        if source is not None:
            question["_source"] = source

        # Call to ES
        if aggregation is not None:
            try:
                question["size"] = 0
                response = self.storage.search(question, auto_scroll=False)[
                    "aggregations"
                ]
            except KeyError as e:
                soil.logger.error("on no! %s", e)
        else:
            try:
                response = list(self.storage.search(question)[0])
            except KeyError as e:
                soil.logger.error("on no! %s", e)

        # Apply processing function if so
        soil.logger.info(lang)
        if fn is not None:
            if fn_args is None:
                fn_args = {}
            if fn in fns:
                response = fns[fn](response, **fn_args)
        return response



""" Module to do the predictions given a model and data points to predict on. """
from soil import modulify
from soil.data_structures.predictions import Predictions

# from soil import logger


@modulify(output_types=lambda *input_types, **args: [Predictions])
def make_predictions(model, unlabeled):
    """Module function to do the predictions given a model and data points to predict on."""

    regressor = model.data
    x_labels = model.metadata["x_labels"]

    d_y = regressor.predict(unlabeled.data[x_labels])

    new_df = unlabeled.data[x_labels].copy()
    new_df["species"] = d_y  # Predicted species for each flower in unknown.csv
    new_df_dict = new_df.to_dict("records")
    predictions = list({"_id": i, **v} for i, v in enumerate(new_df_dict))

    return [Predictions(predictions, metadata={"index": "simple-preds"})]



""" This program runs when new data arrives. """
import logging
import argparse
import soil
from soil.modules.to_es_data_structure import to_es_data_structure

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


def _reset_dbs():

    # We replace all data in each used data structure by an empty lists, efectively reseting data bases.
    data = []

    logging.info("Cleaning index: simple-data")
    data_ref = soil.data(data, metadata={"index": "simple-data", "rewrite": True})
    (res,) = to_es_data_structure(data_ref)
    soil.alias("flowers", res)

    logging.info("Cleaning index: simple-preds")
    data_ref = soil.data(data, metadata={"index": "simple-preds", "rewrite": True})
    (res,) = to_es_data_structure(data_ref)
    soil.alias("preds", res)

    # # This can be added (in this case we have to modify the corresponding test)
    # # Also, modules should admit empty predictors
    # logging.info('Cleaning trained-model')
    # data_ref = soil.data(data, metadata={'rewrite': True})
    # soil.alias('trained-model', data_ref)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    _reset_dbs(**vars(args))



""" Module to train the predictor """
from sklearn.ensemble import RandomForestClassifier
import pandas as pd
from soil import modulify
from soil import logger
from soil.data_structures.random_forest_model import RandomForestModel


@modulify(output_types=lambda *input_types, **args: [RandomForestModel])
def retrain_predictor(data):
    """Trains the predictor with available data"""

    model = RandomForestClassifier(oob_score=True)

    metadata = data.metadata
    x_labels = metadata["columns"]["X_columns"]  # list (metada)
    y_labels = metadata["columns"]["y_columns"]

    # Since Data frame can be defined from an iterator, there is no need to acces data.data
    # df = pd.DataFrame(data.get_data())
    df = pd.DataFrame(data)

    model = model.fit(df[x_labels].values, df[y_labels].values.ravel())

    logger.info("Model oob score: %s", model.oob_score_)
    logger.info("Model error: %s", 1 - model.oob_score_)

    model_metadata = {
        "oob_score": model.oob_score_,
        "error": 1 - model.oob_score_,
        "x_labels": x_labels,
        "y_labels": y_labels,
    }

    return [RandomForestModel(model, model_metadata)]
